/*
 * main.cpp
 *     2009-2010 (c) Paul Malishev <p.malishev@gmail.com>
 *
 */
#include <map>
#include <vector>
#include <list>
#include <queue>
#include <string>
#include <pthread.h>
#include <sched.h>
#include <iostream>
#include <stdlib.h>
#include <signal.h>
#include <linux/input.h>
#include <sys/file.h>
#include <execinfo.h>
#include "timeout/Timeout.hpp"
#include "timeval/timeval.hpp"
#include "global_config.hpp"
#include "powerfail_listener.hpp"
#include "hardware/boards/boards_io.hpp"
#include "collector/collector.hpp"
#include "threads/threads.hpp"
#include "threads/service_thread.hpp"
#include "threads/collector_in_thread.hpp"
#include "threads/collector_out_thread.hpp"
#include "threads/manager_thread.hpp"
#include "threads/archiver_thread.hpp"
#include "threads/reports_thread.hpp"
#include "threads/process_signal_thread.hpp"
#include "threads/events_thread.hpp"
#include "threads/math_channels_thread.hpp"
#include "threads/dump_thread.hpp"
#include "threads/commit_thread.hpp"
//#include "threads/IOWorkerThread.hpp"
//#include "threads/IOCommiterThread.hpp"
//#include "threads/IOInterfaceThread.hpp"      // модуль ответа текущих значений прибора
#include "threads/InterfaceInputsThread.hpp"  // модуль опроса интерфейсных входов
#include "threads/rtc_timer_thread.hpp"
#include "threads/asio_thread.hpp"
#include "configurator.hpp"
#include "gui/ui_MainDlg.hpp"
#include "start_params.hpp"
#include "boost/program_options.hpp"
#include "boost/filesystem.hpp"
#include "hardware/external_storage/ExternalStorageManager.hpp"
#include "hardware/external_storage/ExternalStorageDevice.hpp"
#include "hardware/boot/BootConfig.hpp"

#ifndef GUI_DISABLED
#include "gui/widgets_system/WidgetsEngine.hpp"
#include "threads/interface_thread.hpp"
//#include "gui/ui_StatusProgramDlg.hpp"
#endif

#include "sys_config.hpp"
#include "archiver/archiver.hpp"
#include "reports/manager.hpp"
#include "hardware/rtc/RtcManager.hpp"
#include "hardware/screen/Screen.hpp"
#include "hardware/linux_kernel/psu_signal.hpp"
#include "io_controller/io_fstream.hpp"

#ifdef OS_LINUX
# ifndef HARDWARE_F1771
#   include "proc_parser/proc_parser.hpp"
# endif
#endif

#include <sys/resource.h>
#include "singleton_master/singleton_master.hpp"
#include "console/esc_sequence.hpp"
#include "dictionary/dictionary.hpp"
#include "diagnostics/thread_diagnostics/thread_diagnostics.hpp"

#ifndef GUI_DISABLED
#include "manifest_of_notices.h"
#include "gui/GUIManager.hpp"
#endif

/**
 * @mainpage Многоканальный электронный регистратор Regigraf (Ф1771, Ф1772)
 */

namespace regigraf {
	int debug_n_branch;
}

// [©dsmover: 2014-09-08 Mon 05:16 PM]
// где объявить ее чтобы линковались другие тулзы? пока не ясно
namespace start_params {
	uint32_t pid_of_emulator;
}

/// Функтор для вывода отладочных сообщений модуля IOJournal
class IOFStreamDebug : public io_ctl::DebugMsgHandler {
  public:
    IOFStreamDebug(): io_ctl::DebugMsgHandler() {}
    virtual io_ctl::DebugMsgHandler& operator()(const char *msg) {
      SafeMutex guard(print_mutex);
      using namespace console::ansi_esc;
      std::cout << Sequence(kTextColor, kBlue, kLight)
                << "IO: " << msg
                << Sequence(kReset)
                << std::endl;
      // TODO: сделать запись в журналы
      //logger::misc_info << "IO: " << msg << end;
      return *this;
    }
};

/// Хэш соответствий номер_сигнала->название_сигнала
std::map<int, std::string> signal_map;
/// По этому значению программа будет перезапущена с помощью main.sh
#define MAIN_EXIT_CODE_NORMAL 0
/// Программе не удалось проинициализироваться
#define MAIN_EXIT_CODE_BAD_INIT   5
/// По этому значению прибор будет перезагружен с помощью main.sh
#define MAIN_EXIT_CODE_REBOOT 50
/// По этому значению main.sh завершится. Программа больше не будет перезапускаться
#define MAIN_EXIT_CODE_STOP   51
/// По этому коду значит что случился SegFAULT
#define MAIN_EXIT_SEG_FAULT  101
/// Глобальная переменная, значение которой будет возвращено из main() по return
int mainExitCode = MAIN_EXIT_CODE_NORMAL;

#include <cassert>

/**
 * Сброс логов на устройство хранения и других файлов на устройство хранения.
 *
 * Пробегает по всем объектам логов и вызывает метод flush() у них.
 * Затем вызывает sync(), который скидывает все буферы на диск (флэшку).
 */
static void flush() {
  // сохраним структуры архивов: виртуальных пакетов, событий
  math::Manager::Instance()->Dump();
//  Archiver::Instance()->Flush(); нету больше старого архива и его метода flush..
  logger::LogManager::Instance()->DumpIndeces();
  reports::Manager::Instance()->Dump();
  sync();
}
/**
 * Обработчик сигналов.
 * Пишет в лог информацию. Сохраняет логи. Начинает процесс остановки программы.
 */
volatile static bool segfault_already_happened = false;
static void signal_catcher(int sig_num) {
  struct SegFailException : public std::exception {
    SegFailException(const std::string &_stack) throw(): std::exception(),
                                                         stack(_stack) {
    }
    virtual ~SegFailException() throw() {}
    const char* what() const throw() {
      return ("Segmentation fault signal. Stack:\n" + stack).c_str();
    }
    const std::string stack;
  };
  logger::misc_error << (LogPacketString() << "Главной программой пойман сигнал "
                                           << signal_map[sig_num]);
  using namespace regigraf::sync::queue;
  using namespace regigraf::sync;
	switch (sig_num) {
    case SIGPWR:
      mainExitCode = MAIN_EXIT_CODE_REBOOT;
      break;
    case SIGUSR2:
      mainExitCode = MAIN_EXIT_CODE_STOP;
      break;
    case SIGUSR1:
      // 03.07.2015 denis: для прекращения обработки данных строго по сигналу от
      // эмулятора плат:
      // 1. Дожидаемся завершения цепочки
      // 2. Устанавливаем все цепочки на паузу
      // 3. Освобождаем цепочки, чтоб мог отработать алгоритм "долина смерти",
      //    см. main.cpp:finish()
      regigraf::sync::PhaseManager::WaitForAllChains();
      regigraf::sync::PhaseManager::PauseAllChains();
      regigraf::sync::PhaseManager::ReleaseAllChains();
      mainExitCode = MAIN_EXIT_CODE_NORMAL;
      break;
    case SIGABRT:
    case SIGSEGV: {
      if (segfault_already_happened)
        return;
      static const unsigned kStackBufSz = 100;
      int    i,
             nptrs;
      void  *buf[kStackBufSz];
      char **strings;
      nptrs   = backtrace(buf, kStackBufSz);
      strings = backtrace_symbols(buf, nptrs);
      std::stringstream stack_log;
      if (strings != NULL) {
        for (i = 0; i < nptrs; i++) {
          stack_log << strings[i] << std::endl;
        }
        free(strings);
      } else {
        stack_log << "no symbols";
      }
		  #ifndef GUI_DISABLED
      std::string thread_to_blame("???");
      proc::Manager::GetInfoFrom<proc::Stat>().GetThreadInfo( (unsigned long) GetTId(), &thread_to_blame, 0);
      std::stringstream str;
      str << "\n \n \nФатальная ошибка!!! SEGFAULT в треде: "
          << "'" << thread_to_blame << "'!\n"
          << "\n \nСтек:\n \n";
      WidgetsEngine::PrintLnOnSplashScreen(str.str());
      WidgetsEngine::PrintLnOnSplashScreen(stack_log.str());
      #else
      std::cout << "SegFAULT: " << stack_log.str() << std::endl;
      #endif

      // Программа умерла, сообщим чтобы наверное и эмулятор тоже закрывался в autotest.sh запущенный
	    if (start_params::pid_of_emulator) {
	    	int nres = kill(start_params::pid_of_emulator, SIGUSR1); // 10 работает тоже
	      if (nres != 0) {
	        std::cout << A_RED << "\nnmain.cpp: Can't send signal for 1772_data_emulator!\n" << A_NORMAL << std::endl;
	    	} else {
	    		std::cout << A_WHITE << "\n\n\nmain.cpp: Signal was sent successfully for 1772_data_emulator!"
	    		  << " time: " << time_to_ascii(TimeVal().GetTimeOfDay()) << "\n" << A_NORMAL << std::endl;
	    	}
      }

		  #ifndef GUI_DISABLED
      str.str("");
      str << " \nВремя падения: " << time_to_ascii(TimeVal()) << "\n \n";
      WidgetsEngine::PrintLnOnSplashScreen(str.str());
      #endif

	    // // [©dsmover: 2015-08-13 Thu 10:29 PM] чтобы при функциональных тестах скрипт autotest.sh
	    // // не стопорился
      // exit(MAIN_EXIT_SEG_FAULT);
      Signal::Queue::CreateGlobalHandle(kSignalsMaster)->Push(Signal(SIGNAL_EXIT));
      if (!segfault_already_happened) {
        segfault_already_happened = true;
        throw SegFailException(stack_log.str());
      }
      break;
    }
    case SIGTERM:
    case SIGINT:
    default:
      mainExitCode = MAIN_EXIT_CODE_NORMAL;
      break;
	}
  Signal::Queue::CreateGlobalHandle(kSignalsMaster)->Push(Signal(SIGNAL_EXIT));
}

#ifndef GUI_DISABLED
static void *initGUI_in_separate_thread(void * /*arg*/) {
	WidgetsEngine::InitSystem();
	WidgetsEngine::ShowSplashScreen(Configurator::Regigraf::GetSoftVer());
	return NULL;
}

static void initGUI() {
	initGUI_in_separate_thread(NULL); // (все же вызовем в нашем потоке)
}
#else
static void initGUI() {
	std::cout << "GUI is turned off" << std::endl;
}
#endif
/**
 * Запись в логи интересной информации об окружении, в котором программа запущена.
 */
static void logEnv() {
	logger::misc_info << (LogPacketString() << "Версия программы: "
	                                        << REGIGRAF_VERSION );
	// пока что все :)
}
/**
 * Регистрация сигналов.
 * Регистрирует свой обработчик на SIGINT, SIGTERM, SIGPWR, SIGUSR1, SIGUSR2.
 */
static void register_signals() {
/// Макрос регистрации сигнала. Сохраняет в хэш название и регистрирует сигнал.
#define REGISTER_SYSTEM_SIGNAL(x) signal_map[x] = #x; signal(x, signal_catcher);
	REGISTER_SYSTEM_SIGNAL(SIGINT);
	REGISTER_SYSTEM_SIGNAL(SIGTERM);
	REGISTER_SYSTEM_SIGNAL(SIGPWR);
	REGISTER_SYSTEM_SIGNAL(SIGUSR1);
	REGISTER_SYSTEM_SIGNAL(SIGUSR2);
	REGISTER_SYSTEM_SIGNAL(SIGSEGV);
	REGISTER_SYSTEM_SIGNAL(SIGABRT);
#undef REGISTER_SYSTEM_SIGNAL
}

static bool RestoreConfiguration() {
	Configurator::Regigraf::Init(Cnf); // todo: refact. убрать глобальную переменную

  const std::string kLicenseFile = "./license.xml";
	if (!io_ctl::is_file_exist(kLicenseFile)) {
    Configurator::Control::Instance()->CreateDefaultLicense(kLicenseFile);
  }

	// todo: проверять код возврата и что-то делать
	// сессия не существует какая либо, вызываем LoadConfigRoutine
	//Configurator::Control::Instance()->LoadConfigRoutine(Cnf, Configurator::NORMAL_CONFIG_FN);
  // [©dsmover: 2015-09-07 Mon 07:25 PM]
  Configurator::Session session(SYSTEM_SESSION,
                                Cnf,
                                false,
                                "MainLoadConfig");
  Configurator::Control::Instance()->LoadConfig(
      &session,
      Configurator::Control::kItemAll,
      Configurator::NORMAL_CONFIG_FN);

	return true;
}

static void init_rtc_clocks() {
  std::cout << "main.cpp: init_rtc_clocks() enter" << std::endl;
	try {
    RtcManager::Instance()->Init();
	} catch (...) {
		// добавил обработку этого исключения временно [dsmover: 2013-10-16 Wed 03:04 PM]
	}
  std::cout << "main.cpp: init_rtc_clocks() exit" << std::endl;
}

static void PrepareDirectories() {
  // проверим, что есть каталог, где таблицы матканалов живут, иначе создадим
  // его
  MkDir("./math_tables");
  // директория файлов нашей программы
  MkDir("./var");
  MkDir("./tmp");
  // директория куда будут сохраняться данные сериализирующихся диалогов
  MkDir("./var/gui_data");
  // директория куда будут сохраняться данные диалогов, только данные хранятся
  // пока процесс программы живой.
  boost::filesystem::remove_all("./tmp/gui_data");
  MkDir("./tmp/gui_data");
#ifndef ARM
  MkDir("./tmp/mmc");
#endif
}
/**
 * Инициализация главного треда.
 *
 * Иниацилизация конфигуратора.
 * Регистрация обработчика сигналов (register_signals()).
 * Заполняет хэш BoardsIO.
 */
static bool init() {
	#ifndef GUI_DISABLED
	char strline[] = "------------------------------------------------------------------------------";
	using namespace regigraf::dictionary;
	WidgetsEngine::PrintOnSplashScreen(
	     regigraf::dictionary::Manager::Instance()->GetToken(kStrInitialization));
	#endif
	try {
		// читаем конфиг
//		Configurator::ConfigSessionsManager::Instance();
//		Configurator::Manager::Instance()->Init(Cnf);

    /*
		logger::misc_debug << (LogPacketString() << "Загрузка конфигурации");
		if (!RestoreConfiguration()) {
			std::cout << "main.cpp: init(): cannot RestoreConfiguration()" << std::endl;
			return false;
		}
    */
		// заполняем хэши BoardsIO
		BoardsIO::fill_boards_maps();
		// Denis: а почему создание и удаление ивентов вынесено сюда?
		// dsmover: ты у кого спрашиваешь? вроде код твой
		// 05.07.2012 denis: вот не пиздите, я тут только дописал то что мне было
		// нужно. А когда дописывал заметил, что менеджер ивентов тут создается!
		// И ненадо тыкать мне в лицо git blame`ом, это всё ложь и провокация!
		// 14.10.2014 denis: перенёс в main
//		events::Manager::Instance()->UseOfValuesRestoration(true);


	} catch (std::exception &e) {
#ifndef GUI_DISABLED
		WidgetsEngine::PrintLnOnSplashScreen("");
		WidgetsEngine::PrintLnOnSplashScreen("");
		WidgetsEngine::PrintLnOnSplashScreen(strline);
		WidgetsEngine::PrintLnOnSplashScreen("Исключение в главной точке ПО (std::exception):");
		WidgetsEngine::PrintLnOnSplashScreen(e.what());
		WidgetsEngine::DoErrorPauseSplashScreen();
#endif
		std::cout << "main.cpp: init(): std::exception: " << e.what() << std::endl;
		return false;
	} catch (...) {
#ifndef GUI_DISABLED
		WidgetsEngine::PrintLnOnSplashScreen(""); WidgetsEngine::PrintLnOnSplashScreen("");
		WidgetsEngine::PrintLnOnSplashScreen(strline);
		WidgetsEngine::PrintLnOnSplashScreen("Поймано неизвестное исключение в главной точке ПО!");
		WidgetsEngine::DoErrorPauseSplashScreen();
#endif
		std::cout << "main.cpp: init(): some exception" << std::endl;
		return false;
	}

#ifndef GUI_DISABLED
  //GUIManagerDlg::ParametersOfRegistration params(notices::GatewayOfMessages, NULL, NULL, true, false, true);
  //GUIManagerDlg::RegistrateNoticeListener("Прослушка сообщений", params);

  GUIManagerDlg::RegistrateMessageGateway();
#endif


	return true;
}
/**
 * Переконфигурация PowerfailListener.
 * Читает из конфига имя файла устройства, где будут появляться события EV_PWR.
 */
static void pfl_reconfig() {
#ifdef HARDWARE_F1771
	std::string devname;
	Node main = Cnf["main"];
	main("powerfail_event")(devname);
	PowerfailListener::Instance()->OpenDev(devname.c_str());
#endif
}
/**
 * Знает о сигнале от Config, по которому переконфигурируется (пока только PowerfailListener).
 * Может читать сигнал из /dev/eventX о потере питания (EV_PWR) с помошью PowerfailListener
 *
 * Выход из функции происходит только когда нужно завершить программу.
 */
static bool stop_main_loop = false;
static void PowerSupplyUnitPoll() {

  static MonotonicClock time_start;

  using namespace regigraf;
  using namespace regigraf::sync;
  using namespace regigraf::sync::queue;
#ifdef HARDWARE_F1771
  const int kMsPollInterval = 1000;
  const int kUsPollInterval = kMsPollInterval * 1000;
  bool should_exit = false;
  std::string pw_devname;
  (*Cnf)["main"]("powerfail_event")(pw_devname);
  PowerfailListener::Ref listener(PowerfailListener::Instance());
  listener->OpenDev(pw_devname.c_str());
  while (!should_exit) {
    // ждем сигнала
    if (listener->is_power_event(kMsPollInterval)) {
      // TODO: запись в журнал
      Screen::Instance()->turnOff();
      should_exit = true;
      break;
    } else {
      usleep(kUsPollInterval);
    }
    should_exit |= stop_main_loop;
  }
#else
  PowerSupplyUnit::Ref psu = PowerSupplyUnit::Instance();

  //static int rg = 0;

  while (not stop_main_loop) {
    const PowerSupplyUnit::Signal kSig = psu->WaitForSignal();
    //usleep(50 * 1000); и не надо слипа, быстрее упадет
    //const PowerSupplyUnit::Signal kSig = rg++ % 2 == 0 ? PowerSupplyUnit::kSave : PowerSupplyUnit::kResurrect;// KILLME!!!!!!!!!!!!!!!
    const MonotonicClock kDetected;
    if (kSig == PowerSupplyUnit::kResurrect) {
      continue;
    }
    if (kSig == PowerSupplyUnit::kUnknown) {
      continue;
    }
    logger::user_info << (LogPacketEvent() << logger::event_power_down);
//    BareThread::Manager::Instance()->GoToDeathValley(__FUNCTION__);
//    flush();
    // DEBUG --------------------
    BareThread::Manager::FindThread<ServiceThread>()->SendPowerFailureSignal();
    std::cout << "DEBUG: " << __FUNCTION__ << ": " << __LINE__ << ": "
              << "Отправлен сигнал сохранения даных: " << MonotonicClock().diff_msec(kDetected)
              << " м.сек.; Сигнал: " << (unsigned)kSig << "; "
              << std::endl;
    // --------------------------
    if (kSig == PowerSupplyUnit::kSave) {
      // 04.08.2014 denis: за это время прибор либо отключится либо питание
      // будет восстановлено
      usleep(500000);
      BareThread::Manager::FindThread<ServiceThread>()->SendPowerRestoreSignal();
      // DEBUG --------------------
      std::cout << "DEBUG: " << __FUNCTION__ << ": " << __LINE__ << ": "
                << "Восстановление работы!"
                << std::endl;
      // --------------------------
    }
//    BareThread::Manager::Instance()->GoOutFromDeathValley();
    // PowerSupplyUnit::kReset
  }
#endif
}
/**
 * Обработка сигналов от тредов.
 * Знает о сигнале SIGNAL_EXIT, по которому завершает программу.
 */
static void master_signal_callback(xml_signals_queue_t &queue) {
  xml_signals_queue_t::ConstAuto guard(queue);
  Signal sig;
  while (queue.UnsafePop(&sig)) {
    // парсим сигнал
    switch (sig.code) {
      case SIGNAL_EXIT:
        if (stop_main_loop)
          break;
        logger::misc_info << (LogPacketString() << "Главному процессу пришел сигнал о завершении");
        stop_main_loop = true;
        regigraf::PowerSupplyUnit::Instance()->StopListening();
        break;
      case XML_UNIT_SIG_UPDATE:
        pfl_reconfig();
        break;
      case SIGNAL_UPDATE_MODE:
        logger::misc_error << (LogPacketString() << "Главному процессу пришел сигнал об обновлении (update mode)");
        Configurator::Control::Instance()->setUpdateOfProgram();
        stop_main_loop = true;
        regigraf::PowerSupplyUnit::Instance()->StopListening();
        break;
      default:
        logger::misc_error << (LogPacketString() << "Главному процессу пришел неизвестный сигнал");
        stop_main_loop = true;
        regigraf::PowerSupplyUnit::Instance()->StopListening();
        break;
    }
  }
}

#ifndef GUI_DISABLED
static void ThreadStartingProgress(const BareThread &thread,
                                   uint16_t          number,
                                   uint16_t          total) {
  using namespace regigraf;
  using namespace regigraf::dictionary;
  dictionary::Manager::Ref dict = dictionary::Manager::Instance();
  WidgetsEngine::PrintOnSplashScreen(
      thread.GetName() + ": " + (std::string)(
          thread.isConfigured() ? dict->GetToken(kStrReady) :
                                  dict->GetToken(kStrInitialization)
      ),
      number == 0 ? 0 : ((float)number / total) * 100
  );
}
#endif


static long GeneretePeerIdRoutine() {
  return stream::Manager::Instance()->GeneratePeerId();
}

/**
 * Запуск потоков.
 */
static void Start() {
  using namespace regigraf::sync::queue;
	logger::misc_debug << (LogPacketString() << "Запуск потоков");
	xml_signals_queue_t::Handle q_master = Signal::Queue::CreateGlobalHandle(kSignalsMaster);
	q_master->set_push_callback(master_signal_callback);
  Cnf->regUnit("main", q_master);
	BareThread::Manager::Ref thread_man = BareThread::Manager::Instance();
  thread_man->set_generate_peerid_handler(GeneretePeerIdRoutine);

#ifndef GUI_DISABLED
	thread_man->set_thread_start_handler(ThreadStartingProgress);
#endif
	try {
	  thread_man->CreateThread<ServiceThread>();
	  thread_man->CreateThread<CollectorOutThread>()
	      ->DependOn<ArchiverThread>()
	       .DependOn<ServiceThread>();
				 // .DependOn<CollectorInThread>(); // нельзя CollectorIn пускать раннее CollectorOut!!!
				                                    // collector_out обращается к очереде output которая в collector_in
																						// добавляется!!! уже второй раз натыкаюсь на грабли! [©dsmover: 2014-11-24 Mon 02:09 PM]
//	       .WeakDependency<CollectorInThread>(); // input_output - Что это?
	  thread_man->CreateThread<CollectorInThread>()
	      ->DependOn<RtcTimerThread>()
	       .DependOn<ServiceThread>();
	  thread_man->CreateThread<ProcessSignalThread>()
	      ->DependOn<CollectorInThread>()
	       .WeakDependency<ServiceThread>();
	  thread_man->CreateThread<ManagerThread>()
	      ->DependOn<ProcessSignalThread>()
	       .DependOn<ArchiverThread>()
	       .WeakDependency<ServiceThread>()
	       .WeakDependency<MathChannelsThread>();
	  thread_man->CreateThread<EventsThread>()
	      ->DependOn<ManagerThread>()
	       .DependOn<MathChannelsThread>()
	       .WeakDependency<ServiceThread>();
	  thread_man->CreateThread<ArchiverThread>()
	      ->WeakDependency<ServiceThread>();
    thread_man->CreateThread<ReportsThread>()
        ->WeakDependency<ServiceThread>();
	  thread_man->CreateThread<MathChannelsThread>()
        ->WeakDependency<ServiceThread>();
#ifndef GUI_DISABLED
	  thread_man->CreateThread<DumpThread>()
	      ->DependOn<InterfaceThread>();
#else
	  thread_man->CreateThread<DumpThread>();
#endif
	  thread_man->CreateThread<CommitThread>();
	  //thread_man->CreateThread<IOWorkerThread>()
	  //    ->DependOn<CollectorInThread>()
	  //     .DependOn<ServiceThread>()
	  //     .WeakDependency<ManagerThread>();
	  //thread_man->CreateThread<IOCommiterThread>()
	  //    ->DependOn<IOWorkerThread>();
	  //thread_man->CreateThread<IOInterfaceThread>()
	  //    ->DependOn<IOWorkerThread>();
    thread_man->CreateThread<InterfaceInputsThread>()
        ->DependOn<AsioThread>(); // IOWorkerThread
	  //thread_man->CreateThread<RtcTimerThread>();
	  thread_man->CreateThread<RtcTimerThread>()   // RtcTimerThread зависит от ServiceThread потому что
        ->DependOn<ServiceThread>();             //   автотесты которые у нас есть работают через
                                                 //   RtcFakeEmulator, который работает через tcp/ip,
                                                 //   а EthConfig вроде как должен был
                                                 //   иницилизироваться в ServiceThread где и уарты
    thread_man->CreateThread<AsioThread>()
         ->DependOn<CollectorInThread>()
				 .DependOn<ServiceThread>()
				 .WeakDependency<ManagerThread>();

#ifndef GUI_DISABLED
	  thread_man->CreateThread<InterfaceThread>()
	      ->DependOn<ManagerThread>()
	       .DependOn<ProcessSignalThread>()
	       .DependOn<EventsThread>()
	       .DependOn<MathChannelsThread>()
	       .DependOn<CollectorOutThread>()
	       .WeakDependency<ServiceThread>();
		// блядский DirectFB перехватывает SIGINT. Поэтому повторим регистрацию сигналов
		register_signals();
#endif


    thread_man->FinaliseRegistrationOfThreads();
		// создание основной цепочки обработки данных
		// rtc_thread -> Collector_in [read boards_io] -> ProcessSignal --
		// -> Manager [generate channels from IO] -> ProcessSignal -> Manager --
		// -> Events -> Math -> Manager -> ProcessSignal -> Manager -> Events --
		// -> (Math ->) Manager -> Collector_out
		using namespace regigraf::sync;
		PhaseManager::Instance()->AddChain(phase::kMainChain)
		    // этап 1
		    .AddPhase(thread_man->FindThread<RtcTimerThread>().GetPtr())
		    .AddPhase(thread_man->FindThread<CollectorInThread>().GetPtr(), Phase::kFlagUseDefaultTimeOut)
		    .AddPhase(thread_man->FindThread<ProcessSignalThread>().GetPtr())
		    .AddPhase(thread_man->FindThread<ManagerThread>().GetPtr())
		    // этап 2
		    .AddPhase(thread_man->FindThread<ProcessSignalThread>().GetPtr())
		    .AddPhase(thread_man->FindThread<ManagerThread>().GetPtr())
		    .AddPhase(thread_man->FindThread<EventsThread>().GetPtr())
		    .AddPhase(thread_man->FindThread<MathChannelsThread>().GetPtr())
		    // этап 3
		    .AddPhase(thread_man->FindThread<ManagerThread>().GetPtr())
		    // этап 4
		    .AddPhase(thread_man->FindThread<ProcessSignalThread>().GetPtr())
		    .AddPhase(thread_man->FindThread<ManagerThread>().GetPtr())
		    .AddPhase(thread_man->FindThread<EventsThread>().GetPtr())
		    .AddPhase(thread_man->FindThread<MathChannelsThread>().GetPtr())
		    // этап 5
		    .AddPhase(thread_man->FindThread<ManagerThread>().GetPtr())
		    .AddPhase(thread_man->FindThread<CollectorOutThread>().GetPtr(), Phase::kFlagInstantlyGoTo) // [©dsmover: 2015-09-29 Tue 04:37 PM]
		    // этап 6 (чтение ТХС)
        /*
         * [18 дек. 2015 г.] denis: были случаи зависания на данной фазе, она уже не нужна
         */
		    //.AddPhase(thread_man->FindThread<CollectorInThread>().GetPtr(), Phase::kFlagInstantlyGoTo)
		    .AddPhase(thread_man->FindThread<EventsThread>().GetPtr(),   Phase::kFlagInstantlyGoTo)
		    .AddPhase(thread_man->FindThread<ReportsThread>().GetPtr(),  Phase::kFlagInstantlyGoTo)
		    .AddPhase(thread_man->FindThread<ArchiverThread>().GetPtr(), Phase::kFlagInstantlyGoTo)
		    .Finalise(RtcManager::GetGlobalPollPeriod());
				// пуск тредов

    // То есть короче, получается что треды цепочки этих фаз есть:
    // 1) RtcTimerThread
    // 2) CollectorIn
    // 3) ProcessSignal
    // 4) ManagerThread
    // 5) EventsThread
    // 6) MathThread
    // 7) CollectorOut
    // 8) Reports
    // 9) Archiver

		xmlSignalsManager::Instance()->send_signal(Signal(XML_UNIT_SIG_UPDATE), NULL);
		Signal::Queue::CreateGlobalHandle(kSignalsMaster);
		thread_man->StartAllThreads();
	} catch (std::exception &e) {
		std::cerr << "main(): не могу инициализировать потоки: поймано std::exception: " << e.what() << std::endl;
		exit(1);
// 18.11.2014 denis: не смог заставить отображаться экран смерти
//		BareThread::Manager::Instance()->ShowDeathScreen(e.what());
	} catch (...) {
		std::cerr << "main(): не могу инициализировать потоки: поймано неизвестное исключение!" << std::endl;
		exit(1);
// 18.11.2014 denis: не смог заставить отображаться экран смерти
//		BareThread::Manager::Instance()->ShowDeathScreen("неизвестное исключение!");
	}
	logger::misc_debug << (LogPacketString() << "Потоки запущены");
	//Configurator::Manager::Instance()->StartupProgramComplete();
#ifndef GUI_DISABLED
	// TODO: может стоит врубать генерацию данных тута?
	WidgetsEngine::CloseSplashScreen();
	ProgramFullStartMes param;
	Configurator::Session::SendMsgToGUI(MAIN_DLG_ID, &param, 30000);
	notices::ManagerOfNotices::Instance()->SetMarkThatProgramStarted();
#endif
	Configurator::Manager::Instance()->StartupProgramComplete();
	if (start_params::pid_of_emulator) {
		int nres = kill(start_params::pid_of_emulator, SIGUSR1); // 10 работает тоже
	  if (nres != 0) {
	    std::cout << A_RED << "\nnmain.cpp: Can't send signal for 1772_data_emulator!\n" << A_NORMAL << std::endl;
		} else {
			std::cout << A_WHITE << "\n\n\nmain.cpp: Signal was sent successfully for 1772_data_emulator!"
			  << " time: " << time_to_ascii(TimeVal().GetTimeOfDay()) << "\n" << A_NORMAL << std::endl;
		}
  }
}
/**
 * Остановка программы.
 *
 * Сбрасывает логи.
 * Рассылает всем тредам сигнал остановки.
 * Ждет выхода всех тредов, кроме service (с логами).
 * После выхода всех тредов еще раз сохраняет логи и убивает сервисный поток.
 */
static void finish() {
  const bool update_mode = Configurator::Control::Instance()->IsProgramBeingUpdated();
  std::cout << "main.cpp: finish() enter. update_mode=" << update_mode << std::endl;
	logger::user_info << (LogPacketEvent() << logger::event_turnOn << logger::PktEvents(0, 2));
	// Дождёмся завершения работы цепочки обработки данных
	using namespace regigraf::sync;
	BareThread::Manager::Instance()->GoToDeathValley(__FUNCTION__);
	// Выключим генерацию данных
  PhaseManager::PauseAllChains();
	// запускаем цепочку, для коректного завершения работы всех тредов
	BareThread::Manager::Instance()->GoOutFromDeathValley();
  // рассылаем всем сигнал о выходе
  logger::misc_debug << (LogPacketString() << "Рассылка сигналов на остановку потоков");
  logger::user_info  << (LogPacketEvent()  << logger::event_turnOn << logger::PktEvents(0, 3));
  // завершить все треды (возможно кроме ГУИ треда, так как активен режим обновления программы)
  BareThread::Manager::Instance()->StopAllThreads(update_mode ? false : true);
  std::cout << "main.cpp: finish(): threads stopped" << std::endl;
	// отправка сигнала эмулятору плат
	if (start_params::pid_of_emulator) {
		int nres = kill(start_params::pid_of_emulator, SIGUSR2);
	  if (nres != 0) {
	    std::cout << A_RED
	              << "\n\n\n\nmain.cpp: Warning! Remote program wasn't notifyed of ending of the program.\n"
			          << "Couldn't send a signal!" << A_NORMAL << std::endl;
		} else {
	    std::cout << A_WHITE
	              << "\nmain.cpp: remote program was notifyed that program had ended\n"
					      << A_NORMAL << std::endl;
		}
  }

  if (update_mode) {
    std::cout << "main.cpp: finish() let's call Control::ReadyToUpdate()" << std::endl;
    Configurator::Control::Instance()->ReadyToUpdate();
    std::cout << "main.cpp: finish() let's call Control::ReadyToUpdate() returned" << std::endl;
    // режим обновления программы. дождемся фоновой обработки
    bool update_ok = Configurator::Control::Instance()->WaitEndOfUpdateSec(120); // не больше 120 секунд
    std::cout << "update has probably finished. at least update_ok=" << update_ok << std::endl;
    BareThread::Manager::Instance()->StopUpdateThreads();
  }
}
/**
 * Деинициализиация.
 */
static void Deinit() {
  // 15.10.2014 denis: последовательность удаления синглтонов очень важна!
  // 1. Первыми убиваем треды, которые могут содержать ссылки на синглтоны.
  // 2. Удаляем все синглтоны
  // ---------------------------------------------------------------------------
  // удаляем все треды
  BareThread::Manager::Instance()->DestroyAllThreads();

  // удаляем цепочки фаз
  using namespace regigraf::sync;
  PhaseManager::DropAllChains();
  // удаляем все синглтоны
  /// TODO: Выпилить к хуям глобальный Cnf!!!!
  Cnf.Free();
  regigraf::singleton::Master::Destroy();
}

#ifdef CHECK_CRC_ON_STARTUP
static crc::crc32_val_t GetEXECrc(const char *filename) {
	crc::crc32       our_crc;
	crc::crc32_val_t calced;
	FILE *pFile;
	pFile = fopen(filename, "r");
	if (!pFile) {
		std::cerr << "Cannot open the main file to caclulate its CRC sum\n";
		return -1;
	}
	char buf[65536];
	int r;
	while (!feof(pFile)) {
		r = fread(buf, 1, sizeof(buf), pFile);
		for (int i = 0; i < r; i++) {
			our_crc << buf[i];
		}
	}
	fclose(pFile);
	calced = our_crc.calculate();
	return calced;
}
#endif

#ifdef CHECK_CRC_ON_STARTUP
// return: true  - if should restart
//      or false - if not
bool CheckIntegrityOfProgram()
{
	// Посчитаем и запомним контрольную сумму программы.
	int nres;
	unsigned long t1, t2;
	t1 = GetTickCount();
	crc::crc32_val_t crc = GetEXECrc("./main");
	t2 = GetTickCount();
	printf("main(): crc: %X; time calculation: %ldms\n", crc, t2 - t1);
	CONFIG_SYSTEM_SET_VALUE(last_crc, crc);
	// Не совпала! но может в бекапах, что целое лежит
	if (crc != regigraf::kDefaultMainCRC) {
#ifndef GUI_DISABLED
		WidgetsEngine::PrintLnOnSplashScreen("--------------------------------------------------------------------");
		WidgetsEngine::PrintLnOnSplashScreen(" * Внимание! Контрольная сумма не совпала!!! Программа повреждена! *");
		WidgetsEngine::PrintLnOnSplashScreen("--------------------------------------------------------------------");
#endif
		
		if (io_ctl::is_file_exist("./backup/main")) {
			crc = GetEXECrc("./backup/main");
			if (crc == regigraf::kDefaultMainCRC) {
				#ifndef GUI_DISABLED
				WidgetsEngine::PrintLnOnSplashScreen("Идет восстановление рабочей версии из бепака...");
				#endif
				nres = system("cp ./backup/main ./main");
				(void) nres;
#ifndef GUI_DISABLED
				WidgetsEngine::PrintLnOnSplashScreen("Скопировано! Программа перезапустится через 10 секунд.");
#endif
				// только сделаем чтобы при старте программа не обновляла папку бекапа,
				// а то получится бекапим, то из чего восстановились
				nres = system("touch ./main ./backup/main");
				(void) nres;
				
				chmod("./main", S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IWGRP |
				                S_IXGRP | S_IROTH | S_IWOTH | S_IXOTH);
				
				sleep(10);
				return true;
			}
		}
		
		// Запускаться, значит, нет с чего: плохо - фатальная ошибка!
		#ifndef GUI_DISABLED
		WidgetsEngine::PrintLnOnSplashScreen("ФАТАЛЬНАЯ ОШИБКА! НЕ УДАЕТСЯ НАЙТИ ГЛАВНЫЙ МОДУЛЬ ИСПОЛНЕНИЯ ПРОГРАММЫ!");
		#endif
		sleep(30);
		return true;
	}

	// Все нормально, все запустилось (запустится :), проверим, может бекапчик пора обновить
	if (!io_ctl::is_file_exist("./backup")) {
		// Надо папку создать
		mkdir("./backup", 0777);
		
	}
	//chmod("./backup", S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IWGRP | S_IXGRP | S_IROTH | S_IWOTH | S_IXOTH);
	
	bool ought_to_make_backup = false;
	
	if (!io_ctl::is_file_exist("./backup/main")) {
		ought_to_make_backup = true;
	} else {
		struct stat stat_orig, stat_bckup;
		stat("./main",        &stat_orig);
		stat("./backup/main", &stat_bckup);
		if (stat_orig.st_size != stat_bckup.st_size   ||    // размеры файлов
		    stat_orig.st_mtime != stat_bckup.st_mtime ||    // время последней модификации
		    stat_orig.st_ctime != stat_bckup.st_ctime)      // время последней смены статуса файла
		  ought_to_make_backup = true;
	}
	
	if (ought_to_make_backup) {
		#ifndef GUI_DISABLED
		WidgetsEngine::PrintLnOnSplashScreen("");
		WidgetsEngine::PrintLnOnSplashScreen("Подождите! Идет создание бекапа рабочей версии!");
		#endif
		nres = system("cp ./main ./backup/main");
		(void) nres;
		nres = system("touch ./main ./backup/main");
		(void) nres;
		
		//chmod("./main", S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IWGRP | S_IXGRP | S_IROTH | S_IWOTH | S_IXOTH);
		
		#ifndef GUI_DISABLED
		WidgetsEngine::PrintLnOnSplashScreen("--- Бекап готов ----");
		#endif
		
		// После бекапа все равно надо перезапустить программу из-за вызова "touch ./main ./backup/main"
		return true;
	}
	
	nres = system("chmod a+rwx . ./main backup ./backup/main");
	(void) nres;
	
	return false;
}
#endif

#ifndef GUI_DISABLED
extern unsigned long events_check_correct_expression;
#endif

void PrintEnterUpdateAllEntitiesWhoIam();
void PrintEnterUpdateAllEntitiesWhoIam() {
  std::string thread_to_blame("???");
  proc::Manager::GetInfoFrom<proc::Stat>().GetThreadInfo( (unsigned long) GetTId(), &thread_to_blame, 0);

  std::cout << "UpdateAllEntities() enter. thread is '" << thread_to_blame << "'" << std::endl;
}

void PrintExitUpdateAllEntitiesWhoIam();
void PrintExitUpdateAllEntitiesWhoIam() {
  std::string thread_to_blame("???");
  proc::Manager::GetInfoFrom<proc::Stat>().GetThreadInfo( (unsigned long) GetTId(), &thread_to_blame, 0);

  std::cout << "UpdateAllEntities() exit. thread is '" << thread_to_blame << "'" << std::endl;
}

namespace start_params {
//std::string boards_io_input_file = "";
std::string tcp_data_server;
int         tcp_base_port_server;
}

static void ReadBootConfig(const boost::program_options::variables_map &vm) {
  BootConfig boot_conf;
  boot_conf.AddBootOptions("lcd_model", vm["lcd_model"].as<std::string>());
  Screen::Instance()->set_model(boot_conf["lcd_model"]);
}

static void PrintBuildInfo() {
  std::cout << "Software version: " << Configurator::Regigraf::GetSoftVer() << ";\n"
            << "Builder         : " << REGIGRAF_BUILDER << ";\n"
            << "Date of build   : " << __DATE__ << " " << __TIME__ << ";\n"
            << "Commit          : " << REGIGRAF_COMMIT << ";\n"
            << "Date of commit  : " << REGIGRAF_COMMIT_DATE << ";"
            << std::endl;
}

//функция для поиска процесса по имени
pid_t proc_find(const char* name) {
	DIR* dir;
	struct dirent* ent;
	char* endptr;
	char buf[512];
	if (!(dir = opendir("/proc"))) {
			perror("can't open /proc");
			return -1;
	}
	while((ent = readdir(dir)) != NULL) {
			/* if endptr is not a null character, the directory is not
			 * entirely numeric, so ignore it */
			long lpid = strtol(ent->d_name, &endptr, 10);
			if (*endptr != '\0') {
					continue;
			}
			/* try to open the cmdline file */
			snprintf(buf, sizeof(buf), "/proc/%ld/cmdline", lpid);
			FILE* fp = fopen(buf, "r");
			if (fp) {
					if (fgets(buf, sizeof(buf), fp) != NULL) {
							/* check the first token in the file, the program name */
							char* first = strtok(buf, " ");
							if (!strcmp(first, name)) {
									fclose(fp);
									closedir(dir);
									return (pid_t)lpid;
							}
					}
					fclose(fp);
			}
	}
	closedir(dir);
	return -1;
}

/**
 * Главная программа.
 *
 * Инициализируется (init()), запускает треды (start()).
 * Затем ждет сигналов от тредов (signal_parser()).
 * После получения сигнала выхода завершает программу (finish(), deinit()).
 */
int main(int argc, char *argv[]) {
#ifdef HARDWARE_F1771
  (void) argc;
  (void) argv;
#else

//вызываем журналер, пока тут, но потом возможно перенесём. +если журналер уже открыт внешне, то запускать его не надо.
  int pid;
  pid_t pid1 = proc_find("./journaler");
  pid_t pid2 = proc_find("journaler");
  if ((pid1 == -1) && (pid2 == -1)) {
      printf("journaler not found\n");
      pid = fork();
      if(!pid){
      execlp("./journaler","journaler","p", (char *)NULL);
      }
  } else {
  	if (pid1 != -1) {
      printf("journaler pid = %d\n", pid1);
      pid = pid1;
  	}
  	if (pid2 != -1) {
      printf("journaler pid = %d\n", pid2);
      pid = pid2;
  	}
  }

  int pid_j;
  pid_t pid_j1 = proc_find("./archiver");
  pid_t pid_j2 = proc_find("archiver");
  if ((pid_j1 == -1) && (pid_j2 == -1)) {
      printf("archiver not found\n");
      pid_j = fork();
      if(!pid_j){
      execlp("./archiver","archiver","p", (char *)NULL);
      }
  } else {
  	if (pid_j1 != -1) {
      printf("archiver pid_j = %d\n", pid_j1);
      pid_j = pid_j1;
  	}
  	if (pid_j2 != -1) {
      printf("archiver pid_j = %d\n", pid_j2);
      pid_j = pid_j2;
  	}
  }

//почему-то отрабатывал некорректно, разобраться
//  pid_t pid = proc_find("./journaler");
//  if (pid == -1)
//  	pid = proc_find("journaler");
//  printf("./journaler not found\n");
//  if (pid == -1) {
//  	printf("journaler not found\n");
//  }
//    pid = fork();
//    if(!pid){
//    execlp("./journaler","journaler","p", (char *)NULL);
//  }

//  int pid = fork();
//  if(!pid){
//  execlp("./journaler","journaler","p", (char *)NULL);
//  }

	namespace po   = boost::program_options;
	namespace dict = regigraf::dictionary;
	po::options_description desc("This is the regigraf program. There are some keys");
	desc.add_options()
		("help"                       , "produce this help message")
    ("version"                    , po::value<bool>()->default_value(false)->zero_tokens(),
                                          "a flag, print information about program version and immediately stop it")
		("board"                      , po::value<std::string>()->default_value("1772"),
		                                          "possible values are: 1771, 1772")
		("rtc"                        , po::value<std::string>()->default_value("AUTO"),
		                                      "sets what RTC to use: AUTO, FAKE, FAKE_EMULATOR")
	  ("period"                     , po::value<int>()->default_value(100),
		                                      "period in ms (1ms = 1/1000 sec). 100ms is default")
		("tcp_data_server"            , po::value<std::string>()->default_value(""),
                                          "\"\" - switched off, 127.0.0.1 or dns name")
		("tcp_base_port_server"       , po::value<int>()->default_value(4000),
		                                      "tcp ip port. each board is addressed like: base, base+1, base+2")
		("remote_pid"                 , po::value<uint32_t>()->default_value(0),
		                                      "PID of 1772_data_emulator for remote control")
    ("file_test_path_points"      , po::value<std::string>()->default_value("none"),
                                              "file with path points")
    ("time_interval_between_points", po::value<uint32_t>()->default_value(0),
                                          "Time interval between points")
    ("number_of_runs_points", po::value<uint32_t>()->default_value(1),
                                          "The number of runs points")
    ("lcd_model"                   , po::value<std::string>()->default_value(BootConfig::kDefLcdModel),
                                             "possible values are: DEFAULT, G150XGE-L04, NLB150XG01L-01, AA104XF02 and etc.");

		//("file_source_of_data", po::value<std::string>(), "change a source of data (analog boards or file input)");
	po::variables_map vm;
	po::store(po::parse_command_line(argc, argv, desc), vm);
	po::notify(vm);
	if (vm.count("help")) {
		std::cout << desc << "\n";
		return 1;
	}
	if (vm["version"].as<bool>()) {
	  PrintBuildInfo();
	  return 0;
	}
  std::string sboard = vm["board"].as<std::string>();
  if (sboard == "1771") {
    start_params::board = start_params::kBoard_1771;
    std::cout << "Parsing keys... board is set as 1771" << std::endl;
  } else if (sboard == "1772") {
    start_params::board = start_params::kBoard_1772;
    std::cout << "Parsing keys... board is set as 1772" << std::endl;
  } else {
    std::cout << "Parsing keys... warning! Cannot recognize the board value. It gets 1771" << std::endl;
  }
  std::string rtc_s = vm["rtc"].as<std::string>();
  if (rtc_s == "AUTO") {
    start_params::rtc = start_params::kAutoDetect;
    std::cout << "Parsing keys... rtc is set as AUTO" << std::endl;
  } else if (rtc_s == "FAKE") {
    start_params::rtc = start_params::kFake;
    std::cout << "Parsing keys... rtc is set as FAKE" << std::endl;
  } else if (rtc_s == "FAKE_EMULATOR") {
    start_params::rtc = start_params::kFakeEmulator;
    std::cout << "Parsing keys... rtc is set as FAKE_EMULATOR" << std::endl;
  } else {
    start_params::rtc = start_params::kAutoDetect;
    std::cout << "Warning!!! Coudln't guess what rtc type you set! Let it be AUTO then!" << std::endl;
  }

	start_params::period = vm["period"].as<int>();
	std::cout << "Parsing keys... period is equal to " << start_params::period << std::endl;

	start_params::tcp_data_server = vm["tcp_data_server"].as<std::string>();
	start_params::tcp_base_port_server = vm["tcp_base_port_server"].as<int>();
	if (!start_params::tcp_data_server.empty())
		std::cout << "Parsing keys... tcp_data_server is " << start_params::tcp_data_server <<
			". 1772-Emulator of data is on over TCP!\n" << 
			"                tcp_port is " << start_params::tcp_base_port_server << std::endl;
	else
		std::cout << "Parsing keys... tcp_data_server is unset. Regular mode is working." << std::endl;

	start_params::pid_of_emulator = vm["remote_pid"].as<uint32_t>();

	if (start_params::pid_of_emulator) {
		std::cout << "Parsing keys... pid_of_emulator is " << start_params::pid_of_emulator << std::endl;
    RtcManager::Instance()->setPIDofEmulator(start_params::pid_of_emulator);
  }
#ifndef GUI_DISABLED  
  std::string filepoints = vm["file_test_path_points"].as<std::string>();
  if (filepoints != "none") {
    WidgetsEngine::file_with_path_points = filepoints;
    WidgetsEngine::time_interval_between_points = vm["time_interval_between_points"].as<uint32_t>();
    WidgetsEngine::number_of_runs_points = vm["number_of_runs_points"].as<uint32_t>();
  }
#endif // GUI_DISABLED
#endif // HARDWARE_F1771
  // читаем параметры загрузчика
  ReadBootConfig(vm);
  // Программа то запускается в одном экземпляре? Проверим!
  Configurator::Control::Detector instance_detector("/tmp/regigraf.lock");
  if (not instance_detector.IfWasLocked()) {
    std::cout << "\n\nПрограмма уже запущена!\n";
    return -1;
  }
  typedef regigraf::singleton::Master SingletonMaster;
  SingletonMaster::Instance();
  diagnostics::AddStack("Main");
  events::Manager::Instance();
//  Archiver::Instance();
  events::Manager::Instance()->UseOfValuesRestoration(true);
	// Для отладки режимов программы.
	regigraf::debug_n_branch = 0;
	PrepareDirectories();
  // устанавливаем обработчик сигналов
  register_signals();
  // Инициализация специального конфига, для хранения системных переменных
	// Очень важно проинициализировать конфигурацию до инициализации RTC,
	// т.к. из неё читается время последнего перевода часов.
  config::System::Instance()->Init();
  // [©dsmover: 2015-06-10 Wed 01:44 PM]
  if (!RestoreConfiguration()) {
    std::cout << "main.cpp: init(): cannot RestoreConfiguration()" << std::endl;
  }

#ifndef HARDWARE_F1771  
  // Загрузка данных о состояниях и настройках переферии
  proc::Manager::Instance()->UpdateAllEntities();
  proc::Manager::Instance()->SetEnterUpdateAllEntitiesCallback(PrintEnterUpdateAllEntitiesWhoIam);
  proc::Manager::Instance()->SetExitUpdateAllEntitiesCallback(PrintExitUpdateAllEntitiesWhoIam);
  // Установка сетевого интерфейса по умолчанию, можно читать из конфига.
  // Вся последующая работа, в программе, будет производится только с указанным
  // устройством (по умолчанию)!
  // Если данное устройство отсутствует, тогда будет выбрано первое доступное.
  proc::Manager::GetInfoFrom<proc::Net>().SetDefaultDev("eth0");
#endif
  // [©dsmover: 2015-06-10 Wed 01:38 PM]
  // Уже на запуске нужен TCP/IP (для RtcEmulatorFake)
  //UartConfig::Instance()->reconfigure();
  EthConfig::Instance()->reconfigure();
  // Инициализация RTC
  init_rtc_clocks();
#ifndef HARDWARE_F1771  
	proc::Manager::GetInfoFrom<proc::Stat>().SetThreadInfo((unsigned long)GetTId(),
	  "int main() { ... }", getpriority(PRIO_PROCESS, 0));

	// [©dsmover: 2018-01-21 Sun 11:25 PM]
	// let's add regirfb too (this is a web server which shows screenshots of /dev/fb screen
	// and handles mouse clicks and movement
	// actually this process is a part of system (one which eats great amount of cpu) let us not hide it
	// but show it in top, I think it is useful

	proc::Manager::GetInfoFrom<proc::Stat>().AddServiceInfo("regirfb");

	// [©dsmover: 2018-01-21 Sun 11:25 PM]
	// here should be added: new journal service and archive calls (todo)
	// (new journal engine lives in a different process now (as an archiver as well)

  // инициализация списка USB устройств
  usb::Manager::Instance();
#endif
  // Инициализация менеджера хранилищ
  ExternalStorageManager::Instance();
  ExternalStorageDevice::Instance();
	// Проверим контрольную сумму. Конечно, такие проверки делать здесь не очень
	// логично, потому что не безопасно.
	// (Ну запустили мы ПО, и оно запустилось, проверили что ЦРЦ не верно, что
	// теперь делать-то, если мы заведомо знаем, что сейчас работает битый
	// ехе-шник?)
	// TODO: Надо что-то воротить с лаунчером.

	initGUI();
#ifdef CHECK_CRC_ON_STARTUP
	bool should_restart = CheckIntegrityOfProgram();
	if (should_restart) {
#ifndef GUI_DISABLED
		WidgetsEngine::CloseSystem();
#endif
		return MAIN_EXIT_CODE_REBOOT;
	}
#endif
	dict::Manager::Ref dict_man = dict::Manager::Instance();
#ifdef MALLOC_STATISTICS_DEBUG
	if (getenv("LD_PRELOAD") && MallocStatisticModule::TryLoadMallocStatistic()) {
		#ifndef GUI_DISABLED
		WidgetsEngine::PrintLnOnSplashScreen("");
		WidgetsEngine::PrintLnOnSplashScreen(" * Внимание! Активирован режим сбора статистики malloc *");
		WidgetsEngine::PrintLnOnSplashScreen("");
		StatusProgramDlg::malloc_stat_mode = true;
		#endif
	} else {
		// Обычный режим
		#ifndef GUI_DISABLED
		WidgetsEngine::PrintOnSplashScreen(dict_man->GetToken(dict::kStrInitialization));
		#endif
	}
	#else
	// Обычный режим
	#ifndef GUI_DISABLED
	WidgetsEngine::PrintOnSplashScreen(dict_man->GetToken(dict::kStrInitialization));
	#endif
#endif
	io_ctl::debug_msg_handler_ptr = boost::shared_ptr<io_ctl::DebugMsgHandler>(new IOFStreamDebug());
	logger::misc_info << (LogPacketString()
	                  << "-----------------------------------------------------");
	logger::user_info << (LogPacketEvent() << logger::event_turnOn
	                                       << logger::PktEvents(0, 0));
	logEnv();
	// инициализация
	if (!init()) {
	  return MAIN_EXIT_CODE_BAD_INIT;
	}

	// запускаем треды
	Start();
	// Загрузка значения яркости подсветки дисплея из системной конфигурации
	Screen::Instance()->set_brightness(CONFIG_SYSTEM_GET_VALUE(uint8_t, brightness));
	// Основной цикл, в нём опрашивается состояние блока питания прибора!
	PowerSupplyUnitPoll();
	// выходим из программы
	finish();
#ifndef GUI_DISABLED
	if (Widget::count_created_widgets == Widget::count_destructed_widgets) {
		std::cout << "main.cpp: count_created_widgets = " << Widget::count_created_widgets
		     << "; count_destructed_widgets = " << Widget::count_destructed_widgets
				 << std::endl;
	} else {
		std::cout << A_RED << "main.cpp: count_created_widgets = " << Widget::count_created_widgets
		     << "; count_destructed_widgets = " << Widget::count_destructed_widgets
				 << A_NORMAL << std::endl;
	}
#endif
#ifdef MALLOC_STATISTICS_DEBUG
	if (getenv("LD_PRELOAD")) {
	  MallocStatisticModule::UnloadMallocStatistic();
	}
#endif

  if (Configurator::Control::Instance()->IsNeededHardReboot())
    mainExitCode = MAIN_EXIT_CODE_REBOOT;

  // деинициализация
  dict_man.Free();
  Deinit();
  
	return mainExitCode;
}

